# -*- coding: utf8 -*-
'''
DB MODULE - FoxPro ODBC
*******************************
***  DB MODULE - FoxPro ODBC
*******************************
$RCSfile: dbmod_odbc_foxpro.py,v $
$Revision: 1558 $
$Author: j.holetschek $
$Date: 2018-05-08 13:17:19 +0200 (Di, 08. Mai 2018) $

The pywrapper database module to access foxpro ODBC datasources.
You will need the mxODBC module to connect to a DSN via ODBC.
The FoxPro ODBC driver does not seem to support parameter passing.
Therefore no real unicode can be passed and all parameters are encoded first into the PSF setting.
'''


import os
from string                             import join
from biocase.wrapper.errorclasses       import *
from biocase.wrapper.sql.base           import sqlBase
from biocase.wrapper.dbmod.base_dbmod   import DBBasemodule

import logging
log = logging.getLogger("pywrapper.dbmod")


# uses inherited ANSI SQL apart from this:
#
# no IN operator supported
sqlBase.multipleCOP['in'] = None
sqlBase.templates['select-limit'] = 'SELECT DISTINCT TOP #3 #1 FROM #2'
#sqlBase.templates['attribute'] = '[#1].[#2]'
#sqlBase.templates['table'] = '[#1]'

__syntax__ = '''SELECT [ALL | DISTINCT] [TOP nExpr [PERCENT]] [Alias.] Select_Item 
   [[AS] Column_Name] [, [Alias.] Select_Item [[AS] Column_Name] ...] 
   FROM [FORCE] [DatabaseName!] Table [[AS] Local_Alias] 
   [ [INNER | LEFT [OUTER] | RIGHT [OUTER] | FULL [OUTER] JOIN DatabaseName!] 
      Table [[AS] Local_Alias] [ON JoinCondition ...]
      [[INTO Destination] | [TO FILE FileName [ADDITIVE] | TO PRINTER [PROMPT] | TO SCREEN]] 
      [PREFERENCE PreferenceName] [NOCONSOLE] [PLAIN] [NOWAIT]
      [WHERE JoinCondition [AND JoinCondition ...] [AND | OR FilterCondition [AND | OR FilterCondition ...]]]
   [GROUP BY GroupColumn [, GroupColumn ...]] [HAVING FilterCondition] [UNION [ALL] SELECTCommand] 
   [ORDER BY Order_Item [ASC | DESC] [, Order_Item [ASC | DESC] ...]]
'''     

# ---------------------------------------------------------------------------------
# ---------------------------------------------------------------------------------
# ---------------------------------------------------------------------------------
class DBmodule(DBBasemodule):
    __version__ = ''
    __dbmod__ = u'Foxpro using '

    

    # _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _
    def __init__(self, psfObj):
                # call createDBconnection method
        DBBasemodule.__init__(self, psfObj)        

    # ------------------   FUNCTION OVERLOAD OF BASE DBMOD MODULE   ------------------
    # --------------------------------------------------------------------------------
    def createDBconnection(self):
        try:
            import pyodbc as odbc
            self.__class__.__dbmod__ += "pyODBC"
        except:
            try:
                import ceODBC as odbc
                self.__class__.__dbmod__ += "ceODBC"
            except:
                raise DbModuleMissingError('ceODBC/pyODBC')
        loginParas = "DSN=%s;UID=%s;PWD=%s" % (self.psfObj.database, self.psfObj.user ,self.psfObj.password )
        conn = odbc.connect( loginParas )        
        return conn
       
   
    # _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _
    def execDistinct(self, selectList, whereObj=None):
        '''Select distinct statement. The default routine does a regular select and filters only unique records via Python.'''
        # build SELECT clause
        select = "DISTINCT %s" %(join( [self.joinDBAttrObj(dbt) for dbt in selectList], ', ' ) )
        # get SQL string
        (sql, paras) = self.select(selectList, whereObj, selectReplacementString=select)
        
        # execute statement        
        #return [r[0] for r in self.execSQLnonUnicode(sql, [str(p).encode("ascii") for p in paras])]
        return [r[0] for r in self.execSQLnonUnicode(sql, self.deunicodeParams(paras))]
    
    # _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _
    def execSQL(self, sql, paras=None):
        #self.execSQLnonUnicode(sql, [str(p).encode("ascii") for p in paras])
        self.execSQLnonUnicode(sql, self.deunicodeParams(paras))
    # _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _
    def execSelect(self, selectList, whereObj=None, orderByList=[], limit=None):
        # get SQL string
        (sql, paras) = self.select(selectList, whereObj, limit=limit, orderByList=orderByList)

        # execute statement
        #return self.execSQLnonUnicode(sql, [str(p).encode("ascii") for p in paras])
        return self.execSQLnonUnicode(sql, self.deunicodeParams(paras))
    # _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _
    def getTables(self):
        tables = []
        obj = self.conn if self.__dbmod__.endswith('ceODBC') else self.conn.cursor()
        for t in obj.tables():
            if str(t[3]) == 'TABLE' or str(t[3]) == 'VIEW':
                tables.append(str(t[2]))
        return tables
    # _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _
    def getColumns(self, _table):
        try:
            obj = self.conn if self.__dbmod__.endswith('ceODBC') else self.conn.cursor()
            ret = [str(c[3]) for c in obj.columns(catalog=None, schema=None, table=_table)]
            if ret == None:
                return []
            return ret
        except:
            log.error("error fetching columns")
            return []

    # _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _
    def deunicodeParams(self, paras):
        if(paras == None or len(paras) == 0):
            return tuple()

        for p in paras:
            stringTypes = ["<type 'str'>", "<type 'unicode'>"]
            plist = []
            for p in list(paras):                
                if stringTypes.count(str(type(p))):
                    plist.append(str(p).encode("ascii"))
                else:
                    plist.append(p)
        return tuple(plist)

	
    def hasMetaInfo(self):
        return True